#ifdef ATARI_ST
/*
 * The ACIA driver for the Atari ST, both MDI and KBD
 */
#include "../h/const.h"
#include "../h/type.h"
#include "../h/com.h"
#include "../h/sgtty.h"

#include "staddr.h"
#include "stacia.h"

#include "tty.h"

/*
 * Translation from keyboard codes into internal (ASCII like) codes
 * These tables represents a US keyboard, so MINIX.IMG as found on
 * the MINIX-ST BOOT floppy works well in the US.
 * The TOS program FIXKEYS.PRG is supplied to replace the keyboard
 * tables compiled into MINIX.IMG on the BOOT floppy, by the proper
 * national version of the keyboard tables used by TOS at the time
 * of running FIXKEYS.PRG.
 *
 * Since these three tables are not fully sufficient to deal with
 * all the differences, some special code is added in this driver
 * to cope with the national combinations with the ALT key. See
 * the routine national() below.
 *
 * Currently there are no provisions for non-ASCII characters.
 * Only the problem of entering the ASCII character set from a
 * variety of different keyboard layout is solved.
 * Non-ASCII characters cause three kinds of problems:
 *  - character codes above 127 loose their 8th bit in the TTY driver,
 *    if not in RAW mode
 *  - application programs do not know what to do with non-ASCII,
 *    and certainly not with the character codes Atari did assign
 *  - only ASCII characters can be displayed on the screen, since the
 *    font tables used by MINIX-ST have only 128 entries.
 */
 
/* Scan codes to ASCII for unshifted keys */
PUBLIC unsigned char keynorm[] = {
/*00*/	   0,0x1B, '1', '2', '3', '4', '5', '6',
/*08*/	 '7', '8', '9', '0', '-', '=','\b','\t',
/*10*/	 'q', 'w', 'e', 'r', 't', 'y', 'u', 'i',
/*18*/	 'o', 'p', '[', ']','\r',   0, 'a', 's',
/*20*/	 'd', 'f', 'g', 'h', 'j', 'k', 'l', ';',
/*28*/	'\'', '`',   0,'\\', 'z', 'x', 'c', 'v',
/*30*/	 'b', 'n', 'm', ',', '.', '/',   0,   0,
/*38*/	   0, ' ',   0,   0,   0,   0,   0,   0,
/*40*/	   0,   0,   0,   0,   0,   0,   0,   0,
/*48*/	   0,   0, '-',   0,   0,   0, '+',   0,
/*50*/	   0,   0,   0,0x7F,   0,   0,   0,   0,
/*58*/	   0,   0,   0,   0,   0,   0,   0,   0,
/*60*/	   0,   0,   0, '(', ')', '/', '*', '7',
/*68*/	 '8', '9', '4', '5', '6', '1', '2', '3',
/*70*/	 '0', '.','\r',   0,   0,   0,   0,   0,
/*78*/	   0,   0,   0,   0,   0,   0,   0,   0
};
 
/* Scan codes to ASCII for shifted keys */
PUBLIC unsigned char keyshft[] = {
/*00*/	   0,0x1B, '!', '@', '#', '$', '%', '^',
/*08*/	 '&', '*', '(', ')', '_', '+','\b','\t',
/*10*/	 'Q', 'W', 'E', 'R', 'T', 'Y', 'U', 'I',
/*18*/	 'O', 'P', '{', '}','\r',   0, 'A', 'S',
/*20*/	 'D', 'F', 'G', 'H', 'J', 'K', 'L', ':',
/*28*/	 '"', '~',   0, '|', 'Z', 'X', 'C', 'V',
/*30*/	 'B', 'N', 'M', '<', '>', '?',   0,   0,
/*38*/	   0, ' ',   0,   0,   0,   0,   0,   0,
/*40*/	   0,   0,   0,   0,   0,   0,   0, '7',
/*48*/	 '8',   0, '-', '4',   0, '6', '+',   0,
/*50*/	 '2',   0, '0',0x7F,   0,   0,   0,   0,
/*58*/	   0,   0,   0,   0,   0,   0,   0,   0,
/*60*/	   0,   0,   0, '(', ')', '/', '*', '7', 
/*68*/	 '8', '9', '4', '5', '6', '1', '2', '3',
/*70*/	 '0', '.','\r',   0,   0,   0,   0,   0,
/*78*/	   0,   0,   0,   0,   0,   0,   0,   0
};
 
/* Scan codes to ASCII for capslocked keys */
PUBLIC unsigned char keycaps[] = {
/*00*/	   0,0x1B, '1', '2', '3', '4', '5', '6',
/*08*/	 '7', '8', '9', '0', '-', '=','\b','\t',
/*10*/	 'Q', 'W', 'E', 'R', 'T', 'Y', 'U', 'I',
/*18*/	 'O', 'P', '[', ']','\r',   0, 'A', 'S',
/*20*/	 'D', 'F', 'G', 'H', 'J', 'K', 'L', ';',
/*28*/	'\'', '`',   0,'\\', 'Z', 'X', 'C', 'V',
/*30*/	 'B', 'N', 'M', ',', '.', '/',   0,   0,
/*38*/	   0, ' ',   0,   0,   0,   0,   0,   0,
/*40*/	   0,   0,   0,   0,   0,   0,   0,   0,
/*48*/	   0,   0, '-',   0,   0,   0, '+',   0,
/*50*/	   0,   0,   0,0x7F,   0,   0,   0,   0,
/*58*/	   0,   0,   0,   0,   0,   0,   0,   0,
/*60*/	   0,   0,   0, '(', ')', '/', '*', '7',
/*68*/	 '8', '9', '4', '5', '6', '1', '2', '3',
/*70*/	 '0', '.','\r',   0,   0,   0,   0,   0,
/*78*/	   0,   0,   0,   0,   0,   0,   0,   0
};

PRIVATE int	shift1, shift2, capslock, control, alt;	
				/* keep track of key statii */
PRIVATE char	kbdbuf[2*MAX_OVERRUN+2];
				/* driver collects chars here */
PRIVATE message	kbdmes;		/* message used for console input chars */
PRIVATE int	repeatkey;	/* character to repeat */
PRIVATE int	repeattic;	/* time to next repeat */

/*===========================================================================*
 *				kbdint					     *
 *===========================================================================*/
PUBLIC kbdint()
{
  register code, make, k;

  k = kbdbuf[0];
  /*
   * There may be multiple keys available. Read them all.
   */
  while (KBD->ac_cs & A_IRQ) {
	code = KBD->ac_da;
/*	printf("kbd: got %x\n", code & 0xFF); */
	/*
	 * The ST's keyboard interrupts twice per key,
	 * once when depressed, once when released.
	 * Filter out the latter, ignoring all but
	 * the shift-type keys.
         */
	make = code & 0x80 ? 0 : 1;	/* 1=depressed, 0=released */
	code &= 0x7F;
	switch (code) {
	case 0x2A:	/* shift key on left */
		shift1 = make; continue;
	case 0x36:	/* shift key on right */
		shift2 = make; continue;
	case 0x1D:	/* control */
		control = make; continue;
	case 0x38:	/* alt key */
		alt = make; continue;
	case 0x3A:	/* caps lock */
		if (make) capslock ^= 1; continue;
	}
	if (make == 0) {
		repeattic = 0;
		continue;
	}
	repeatkey = code;
	repeattic = 24;	/* delay: 24 * 16 msec == 0.4 sec */
	kbdkey(code);
  }
  if (kbdbuf[0] != k) {
	/* Build and send the interrupt message. */
	kbdmes.m_type = TTY_CHAR_INT;
	kbdmes.ADDRESS = kbdbuf;
	interrupt(TTY, &kbdmes);	/* send message to tty task */
  }
}

/*===========================================================================*
 *				kbdkey					     *
 *===========================================================================*/
PRIVATE kbdkey(code)
{
  register c;

  if (shift1 || shift2)
	c = keyshft[code];
  else if (capslock)
	c = keycaps[code];
  else
	c = keynorm[code];
  if (c) {
	if (alt)
		c = national(code, c | 0x80);
	if (control) {
		if (c == 0xFF) reboot(); /* CTRL-ALT-DEL */
		c &= 0x1F;
	}
	/* Check to see if character is XOFF, to stop output. */
	if (
		(tty_struct[CONSOLE].tty_mode & RAW) == 0
		&&
		tty_struct[CONSOLE].tty_xoff == c
	) {
		tty_struct[CONSOLE].tty_inhibited = STOPPED;
		return;
	}
	kbdput(c, CONSOLE);
	return;
  }
  if (control && alt && code >= 0x3B && code <= 0x44) {
	/*
	 * some special sequences
	 */
	kbdput(code - 0x3B + 1, OPERATOR);
	return;
  }
  switch (code) {
  case 0x3B:	/* PF1 */
	kbdpf('P'); return;
  case 0x3C:	/* PF2 */
	kbdpf('Q'); return;
  case 0x3D:	/* PF3 */
	kbdpf('R'); return;
  case 0x3E:	/* PF4 */
	kbdpf('S'); return;
  case 0x3F:	/* PF5 */
	kbdpf('T'); return;
  case 0x40:	/* PF6 */
	kbdpf('U'); return;
  case 0x41:	/* PF7 */
	kbdpf('V'); return;
  case 0x42:	/* PF8 */
	kbdpf('W'); return;
  case 0x43:	/* PF9 */
	kbdpf('X'); return;
  case 0x44:	/* PF10 */
	kbdpf('Y'); return;
  case 0x47:	/* Home/Clr */
	kbdansi('H');
	if (shift1 || shift2)
		kbdansi('J');
	return;
  case 0x48:	/* Up */
	kbdansi('A'); return;
  case 0x4B:	/* Left */
	kbdansi('D'); return;
  case 0x4D:	/* Right */
	kbdansi('C'); return;
  case 0x50:	/* Down */
	kbdansi('B'); return;
  default:
	return;
  }
}

/*
 * Cope with national ALT and SHIFT-ALT combinations.
 * Currently only for Germany, France and Spain.
 * Extension are straightforward.
 */
PRIVATE national(code, c)
{
  static char germany[] = {
	0x1A, '@', '\\',
	0x27, '[', '{',
	0x28, ']', '}',
	0
  };
  static char france[] = {
	0x1A, '[', '{',
	0x1B, ']', '}',
	0x28, '\\', 0,
	0x2B, '@', '~',
	0
  };
  static char spain[] = {
	0x1A, '[', '{',
	0x1B, ']', '}',
	0x28, 0x81, 0x9A,	/* lower/upper u-umlaut */
	0x2B, '#', '@',
	0
  };
  register char *p;

  /*
   * Distinguish the right keyboard version somehow
   */
  if (keynorm[0x1A] == 0x81)	/* lower u-umlaut */
	p = germany;
  else if (keynorm[0x1A] == '^')
	p = france;
  else if (keynorm[0x1A] == '\'')
	p = spain;
  else
	return(c);
  /*
   * See if it is one of the keys that need special attention
   */
  while (*p != code) {
	if (*p == 0)
		return(c);
	p += 3;
  }
  /*
   * It is indeed special. Distinguish between upper and lower case.
   */
  p++;
  if (shift1 || shift2)
	p++;
  if (*p == 0)
	return(c);
  return(*p);
}

/*
 * Store the character in memory
 */
PRIVATE kbdput(c, line)
{
  register k;

  if ((k = kbdbuf[0]) >= kbdbuf[1])
	/*
	 * Too many characters have been buffered.
	 * Discard excess.
	 */
	return;
  /* There is room to store this character; do it. */
  k <<= 1;			/* each entry uses two bytes */
  kbdbuf[k+2] = c;		/* store the char code */
  kbdbuf[k+3] = line;		/* which line it came from */
  kbdbuf[0]++;			/* increment counter */
}

/*
 * Input escape sequence for function keys
 */
PRIVATE kbdpf(c)
{
	kbdput('\033', CONSOLE);
	kbdput('O', CONSOLE);
	kbdput(c, CONSOLE);
}

/*
 * Input ANSI escape sequence
 */
PRIVATE kbdansi(c)
{
	kbdput('\033', CONSOLE);
	kbdput('[', CONSOLE);
	kbdput(c, CONSOLE);
}

/*===========================================================================*
 *				kb_timer				     *
 *===========================================================================*/
PUBLIC kb_timer()
{
  register k, s;

  s = lock();
  if (repeattic == 0) {
	restore(s);
	return;
  }
  if (--repeattic != 0) {
	restore(s);
	return;
  }
  k = kbdbuf[0];
  kbdkey(repeatkey);
  if (kbdbuf[0] != k) {
	/* Build and send the interrupt message. */
	kbdmes.m_type = TTY_CHAR_INT;
	kbdmes.ADDRESS = kbdbuf;
	interrupt(TTY, &kbdmes);	/* send message to tty task */
  }
  repeattic = 4;	/* repeat: 4 * 16 msec == 0.066 sec */
  restore(s);
}

/*===========================================================================*
 *				kbdinit				     	     *
 *===========================================================================*/
PUBLIC kbdinit()
{
  KBD->ac_cs = KBD_INIT | A_RXINT;
	/* divide by 16, 8 data, 1 stop, no parity, enable interrupts */
  KBD->ac_da = 0x12;
	/* mouse off */
  kbdbuf[1] = MAX_OVERRUN;	/* set up limit on keyboard buffering */
}
#endif ATARI_ST
