#ifdef ATARI_ST
/* This file contains the main program of MINIX for the Atari ST.
 * The routine main() initializes the system and starts the ball
 * rolling by setting up the proc table, interrupt vectors, and
 * scheduling each task to run to initialize itself.
 * 
 * The entries into this file are:
 *   main:		MINIX main program
 *   none:		called for an interrupt to an unused vector
 *   rupt:		called for an unexpected interrupt (async)
 *   trap:		called for an unexpected trap (synchronous)
 *   panic:		abort MINIX due to a fatal error
 */

#include "../h/const.h"
#include "../h/type.h"
#include "../h/callnr.h"
#include "../h/com.h"
#include "../h/error.h"
#include "../h/signal.h"
#include "const.h"
#include "type.h"
#include "glo.h"
#include "proc.h"

#define SIZES              8	/* sizes array has 8 entries */

/*===========================================================================*
 *                                   main                                    * 
 *===========================================================================*/
PUBLIC main()
{
/* Start the ball rolling. */

  register struct proc *rp;
  register int t;
  register vir_clicks size;
  register phys_clicks base;
  extern unsigned sizes[8];	/* table filled in by build */
  extern int (*task[])();

  initst();
  /* Set up proc table entry for user processes. */
  for (rp = &proc[0]; rp <= &proc[NR_TASKS+LOW_USER]; rp++)	/* DEBUG */
	for (t=0; t< NR_REGS; t++) rp->p_reg[t] = 0x0101*t;	/* DEBUG */

  size = sizes[0] + sizes[1];	/* kernel text + data size */
  base = size;			/* end of kernel */

  for (t = -NR_TASKS; t < 0; t++) {	/* for all drivers */
	rp = proc_addr(t);
	rp->p_sp = (int *)&t_stack[(-t)-1][TASK_STACK_BYTES];
	rp->p_splow = rp->p_sp;
	rp->p_pcpsw.pc = task[t + NR_TASKS];
	if (rp->p_pcpsw.pc != (int (*)())0)
		ready(rp);
	rp->p_pcpsw.psw = 0x2200;	/* S-BIT, SPL2 */
	rp->p_map[T].mem_len  = sizes[0];
/*	rp->p_map[T].mem_phys = 0; */
	rp->p_map[D].mem_len  = sizes[1]; 
	rp->p_map[D].mem_phys = sizes[0];
/*	rp->p_map[S].mem_len  = 0; */
	rp->p_map[S].mem_phys = size;
/*	rp->p_map[T].mem_vir  = rp->p_map[T].mem_phys; */
	rp->p_map[D].mem_vir  = rp->p_map[D].mem_phys;
	rp->p_map[S].mem_vir  = rp->p_map[S].mem_phys;
  }

  rp = proc_addr(HARDWARE);
  rp->p_map[D].mem_len  = ~0;	/* maximum size */
  rp->p_map[D].mem_phys = 0;
  rp->p_map[D].mem_vir  = 0;

  for (t = 0; t <= LOW_USER; t++) {
	rp = proc_addr(t);
	ready(rp);
	rp->p_pcpsw.psw = 0x0200;	/* no S-BIT, SPL2 */
	rp->p_pcpsw.pc = (int (*)()) ((long)base << CLICK_SHIFT);
	size = sizes[2*t + 2];
	rp->p_map[T].mem_len  = size;
	rp->p_map[T].mem_phys = base;
	base += size;
	size = sizes[2*t + 3];
	rp->p_map[D].mem_len  = size;
	rp->p_map[D].mem_phys = base;
	base += size;
/*	rp->p_map[S].mem_len  = 0; */
	rp->p_map[S].mem_phys = base;
	rp->p_map[T].mem_vir  = rp->p_map[T].mem_phys;
	rp->p_map[D].mem_vir  = rp->p_map[D].mem_phys;
	rp->p_map[S].mem_vir  = rp->p_map[S].mem_phys;
  }

  for (rp = proc_addr(LOW_USER+1); rp < proc_addr(NR_PROCS); rp++)
	rp->p_flags = P_SLOT_FREE;

  bill_ptr = proc_addr(HARDWARE);	/* it has to point somewhere */
  pick_proc();

  /* go back to assembly code to start running the current process. */
}


/*===========================================================================*
 *                              none, rupt, trap                             * 
 *===========================================================================*/
PUBLIC none()
{
  panic("Nonexisting interrupt. Vector =", proc_ptr->p_trap);
}

PUBLIC rupt()
{
  panic("Unexpected interrupt.  Vector =", proc_ptr->p_trap);
}


PUBLIC trap()
{
  register t;
  register struct proc *rp;
  static char vecsig[] = {
	0, 0, SIGSEGV, SIGBUS, SIGILL, SIGILL, SIGILL, SIGIOT,
	SIGILL, SIGTRAP, SIGEMT, SIGFPE, STACK_FAULT
  };

  rp = proc_ptr;
  t = rp->p_trap;
  if (rp->p_pcpsw.psw & 0x2000)
	panic("trap via vector", t);
  if (t >= 0 && t < sizeof(vecsig)/sizeof(vecsig[0]) && vecsig[t]) {
	t = vecsig[t];
  } else {
	printf("\nUnexpected trap.  Vector = %d\n", t);
	printf("This may be due to accidentally including\n");
	printf("a non-MINIX library routine that is trying to make a system call.\n");
	t = SIGILL;
  }
  if (t != STACK_FAULT) {	/* DEBUG */
	printf("sig=%d to pid=%d at pc=%X\n",
		t, rp->p_pid, rp->p_pcpsw.pc);
	dump();
  }
  cause_sig(cur_proc, t);
}

PUBLIC checksp()
{
  register struct proc *rp;
  register phys_bytes ad;

  rp = proc_ptr;
  if (rp->p_sp < rp->p_splow)
	rp->p_splow = rp->p_sp;
  if (rp->p_map[S].mem_len == 0)
	return;
  ad = (phys_bytes)rp->p_map[S].mem_phys;
  ad <<= CLICK_SHIFT;
  if ((phys_bytes)rp->p_sp > ad)
	return;
  /*
   * Stack violation.
   */
  ad = (phys_bytes)rp->p_map[D].mem_phys;
  ad += (phys_bytes)rp->p_map[D].mem_len;
  ad <<= CLICK_SHIFT;
  if ((phys_bytes)rp->p_sp < ad + CLICK_SIZE)
	printf("Stack low (pid=%d,pc=%X,sp=%X,end=%X)\n",
		rp->p_pid, (long)rp->p_pcpsw.pc,
		(long)rp->p_sp, (long)ad);
  rp->p_trap = 12;	/* fake trap causing STACK_FAULT */
  trap();
}

/*===========================================================================*
 *                                   panic                                   * 
 *===========================================================================*/
PUBLIC panic(s,n)
char *s;
int n; 
{
/* The system has run aground of a fatal error.  Terminate execution.
 * If the panic originated in MM or FS, the string will be empty and the
 * file system already syncked.  If the panic originates in the kernel, we are
 * kind of stuck. 
 */

  if (*s != 0) {
	printf("\nKernel panic: %s",s); 
	if (n != NO_NUM) printf(" %d", n);
	printf("\n");
  }
  dump();
  printf("\nPush RESET button\n");
  for (;;)
	;
}

/*
 * Atari ST specific initialization.
 */

#include "staddr.h"
#include "stacia.h"
#include "stmfp.h"
#include "stsound.h"
#include "stvideo.h"

/*===========================================================================*
 *                                   initst                                   * 
 *===========================================================================*/
PRIVATE initst()
{
  long ad;

  /*
   * both 8-bit ports of the sound chip are configured for output
   */
  SOUND->sd_selr = YM_MFR;
  SOUND->sd_wdat = PA_OUT|PB_OUT;
  /*
   * init port A (Note: low 3 bits inverted)
   */
  SOUND->sd_selr = YM_IOA;
  SOUND->sd_wdat = SOUND->sd_rdat | PA_PSTROBE;
  /*
   * initialize MFP
   */
  MFP->mf_ierb |= (IB_DINT|IB_AINT|IB_TIMC|IB_PBSY);
  MFP->mf_imrb |= (IB_DINT|IB_AINT|IB_TIMC|IB_PBSY);
  /*
   * interrupts to vectors 0x40 to 0x4F
   * automatic end-of-interrupt mode
   */
  MFP->mf_vr = 0x40;
  /*
   * The following code is needed if TOS is not in ROM.
   * It is harmless for more modern systems.
   */
  ad = *((long *)0x042E);	/* TOS variable 'phystop' */
  ad -= 0x8000L;		/* size of VIDEO memory */
  *((long *)0x0436) = ad;	/* TOS variable '_memtop' */
  VIDEO->vd_ramm = (char)(ad >> 8);
  VIDEO->vd_ramh = (char)(ad >> 16);
}

/*===========================================================================*
 *				aciaint					     *
 *===========================================================================*/
PUBLIC aciaint()
{
  if (KBD->ac_cs & A_IRQ)
	kbdint();
  if (MDI->ac_cs & A_IRQ)
	mdiint();
}

/*===========================================================================*
 *				temporary stuff				     *
 *===========================================================================*/
timint(t)	{fake_int("timint", t);}
siaint(t)	{fake_int("siaint", t);}
mdiint(t)	{fake_int("mdiint", t);}
iob(t)		{fake_int("iob", t);}

fake_int(s, t) char *s; {
  printf("Fake interrupt handler for %s. trap = %02x\n", s, t);
}

#ifdef NEEDED
winchester_task()	{fake_task("winchester_task");}

fake_task(s) char *s; {
  message m;

  /* printf("%s alive\n", s); */
  for (;;) {
	receive(ANY, &m);
	printf("%s received %d from %d\n", s, m.m_type, m.m_source);
  }
}
#endif NEEDED

idle() {
  /* printf("idle "); dump(); */
}
#endif ATARI_ST
